/* Lexmark 7000 driver re-written for [currently] the z51.  The other
   models will be re-added as reverse engineering of the relevant
   drivers proceeds. 

   Current driver (C) 2000 Monty, released under GNU GPL license,
   version 2 

   Original lexmark 7000 driver (C) 1999 Henryk Paluch, released under
   GNU GPL license, version 2 
*/

/*
 * cmyk_pbm2z51.c -- convert multiplane cmyk PBMraw-like file into Lexmark z51
 *              printing escape sequences
 *              Usage: ./cmyk_pbm2z51 < file.cmyk_pbm > /dev/lp0
 */

#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <syslog.h>
#include <stdarg.h>

#include "misc.h"

static void _logmsg(const char *fmt,va_list ap) {
  char buf[1026];
  vsnprintf(buf,1024,fmt,ap);
  if(isatty(STDERR_FILENO)){
    fputs(buf,stderr);
    fflush(stderr);
  }else{
    syslog(LOG_INFO,"%s",buf);
  }
}

void logmsg(const char *fmt,...) {
  va_list ap;
  va_start(ap,fmt);
  _logmsg(fmt,ap);
  va_end(ap);
}

void logerror(const char *fmt,...) {
  va_list ap;
  va_start(ap,fmt);
  _logmsg(fmt,ap);
  va_end(ap);
  exit(1);
}

/*  parse_pbm_header() - decodes header of PBM bitmap
 *        sets global vars: height (pixels), 
 *                          bwidth (real bitamp width - bytes)
 *                          width  (memory bitmap width - bytes)
 *        PBMraw is simple monochromatic bitmap. It consists
 *        of text header and raw data bytes.
 *        Example: P4
 *                 # comment lines
 *                 width height
 *                 raw bytes
 *                 
 *                 width & height is in pixels, lines are byte aligned
 *                 0x80 is most left bit, 0x1 most right
 */
int parse_pbm_header(FILE *f,long *width,long *height,long *plates, int mayeof){
   char buf[1024];
   int dontscan=0;
   
   if (fgets(buf,1023,f)==0  || strncmp(buf,"P4",2)!=0){
      if (mayeof)
	 return 0;
      else
	logerror("Illegal pbmraw header");
   }
   *plates=1; /* assume B&W PBM RAW bitmap on input */
   do{
      fgets(buf,1023,f); /* skip comment line(s) */
      if ( strncmp(buf,"#PI",3)==0 ){  /* interlaced colors CMY(K) lines? */
	 *plates=(int)(buf[3]-'0');
	 sscanf(buf+4,"%ld %ld",width,height);
	 dontscan=1;
      }
   }while(buf[0]=='#');

   if (!dontscan)
      sscanf(buf,"%ld %ld\n",width,height);
   if ( (*width) <=0 || (*height) <=0)
     logerror("Illegal width (%ld) or height (%ld)\n",*width,*height);

   return 1;
}

/* read_color_plates() for one pass
 * f - file to read
 * buf - buffer
 * horizontal offset (bytes)
 * width - to be read (bytes)
 * real buffer width (bytes)
 * buffer height (bytes)
 * xheight (lines to read - may be < height at the end of bitmap)
 * pheight - pass height (now 64 for all the time)
 */
void read_color_plates(FILE *f,unsigned char *buf, int bwidth,
		       int height,int plates){
  int i,j,c;
  
  for(i=0;i<height;i++){
    for(j=0;j<plates;j++){
      c=fread(buf+(j*height*bwidth+bwidth*i),1,bwidth,f);
      if (c==-1)
	logerror("read_color_plates: %s\n",strerror(errno));
      if (c<bwidth)
	logerror("read_color_plates: short read\n");
    }
  }
}  

 
/* read_mono_plate() for one pass
 * f - file to read
 * buf - buffer
 * horizontal offset (bytes)
 * width - to be read (bytes)
 * real buffer width (bytes)
 * buffer height (bytes)
 * xheight (lines to read - may be < height at the end of bitmap)
 */
void read_mono_plate(FILE *f,unsigned char *buf, int bwidth, int height){
  int i,c;
  
  for(i=0;i<height;i++){
    c=fread(buf+bwidth*i,1,bwidth,f);
    if (c==-1)
      logerror("read_mono_plate: fread: %s\n",strerror(errno));
    if (c<bwidth)
      logerror("read_mono_plate: short read %d of %d\n",
	      c,bwidth);
  }
}  
   
static void usage(void){
  logmsg("Lexmark z51 printer driver\n\n");
  logmsg("Usage: cmyk_pbm2z51 [ -h ]"
	 " < input.cmyk_pbm > output.prn\n");
  logmsg("       -h\tthis help text\n");
  logmsg("\nColor printing is turned on, when interlaced"
	 " dithered CMY(K) bitmaps\nare produced by"
	 " pnmraw2cmyk_pbm or bitcmyk2cmyk_pbm\n"
	 " - mode is detected reading \"#PI? w h\"\n"
	 "   where: ? = 3 (CMY) or 4 (CMYK)\n" 
	 "          w = width of one dithered color bitmap\n"
	 "          h = height of one dithered color bitmap\n");
  exit(1);
}

int main(int argc, char *argv[]){
   int i,c;
   int pagecount=0;
   int lremain;
 
   int mayeof=0;
   long width,bwidth,height,plates,psize;
   unsigned char *lines=NULL;

   while( (c=getopt(argc, argv, ":m:hv:"))!=-1 )
      switch(c){
	 case 'h':
	    usage();
	    break;
	 default:
	    usage();
      }


   openlog("cmyk_pbm2z51", LOG_PID|LOG_PERROR, LOG_LPR);
   logmsg("$Id$ started\n");

   /* process raw PBM bitmap(s) */
   while(parse_pbm_header(stdin,&width,&height,&plates,mayeof)){
     mayeof=1;
     bwidth= (width+7) /8; /* round width to whole byte */
     logmsg("bwidth (bytes): %d (%d pixels), height (lines): %d\n",
	    bwidth,bwidth*8, height);
     logmsg("width: %d pixels\n",width);
     lremain=height;
     
     lines=calloc(bwidth*4*height,1);
     
     psize=bwidth*height; /* size of one plate */
     
     if (plates==1){ /* Black & White printing mode */
       read_mono_plate(stdin,lines,
		       bwidth, /*bitmap width bytes read per line*/
		       height); /* buffer height */
       lxmgen_print_page(width, height, bwidth,
			 NULL,NULL,NULL,lines,
			 stdout,&h_dpi600,&v_dpi600,LX_z51);
     }
     if (plates>=3){
       char *cbuf=lines,*mbuf=lines+psize,*ybuf=lines+psize*2,
	 *kbuf=lines+psize*3;
       
       read_color_plates(stdin,lines,
			 bwidth, /*bitmap width bytes read per line*/
			 height, /* lines to be read <=passheight */
			 plates);
       
#if 0
       for(i=0;i<height*bwidth;i++){
	 int black=cbuf[i] & mbuf[i] & ybuf[i];
	 kbuf[i]|=black;
	 cbuf[i]&=~black;
	 mbuf[i]&=~black;
	 ybuf[i]&=~black;
       }
#endif

       lxmgen_print_page(width, height, bwidth,
			 cbuf,mbuf,ybuf,kbuf,
			 stdout,&h_dpi600,&v_dpi600,LX_z51);
       
     } 
     
     pagecount++;
     if (lines!=NULL){
       free(lines);
       lines=NULL;
     }
     
   }
   
   logmsg("printed (%d) page(s)\n",pagecount);
   closelog();
   return 0;
}
